#pylint: disable=logging-format-interpolation
import logging
from datetime import timedelta
from django.conf import settings
from ipware.ip import get_real_ip
from j2fa.errors import TwoFactorAuthError
from j2fa.forms import TwoFactorForm
from j2fa.models import TwoFactorSession
from j2fa.helpers import j2fa_make_code, j2fa_phone_filter
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.http import HttpRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.views.generic import TemplateView


DEFAULT_MAX_FAILED_ATTEMPTS_24H = 5  # can be overriden with settings.MAX_FAILED_ATTEMPTS_24H

logger = logging.getLogger(__name__)


class TwoFactorAuth(TemplateView):
    template_name = 'j2fa/ask-code2.html'
    logout_view_name = 'admin:logout'
    default_next_view_name = 'admin:index'

    def get_user_phone(self, user: User):
        assert user
        assert user.profile
        return user.profile.phone

    def get_context_data(self, **kw):
        request = self.request
        assert isinstance(request, HttpRequest)

        next_url = request.POST.get('next') if request.POST else None
        if not next_url:
            next_url = request.GET.get('next')
        if not next_url:
            next_url = request.META.get('HTTP_REFERER')
        if not next_url:
            next_url = reverse(self.default_next_view_name)

        cx = {
            'form': TwoFactorForm(data=request.POST or None),
            'next': next_url,
        }
        for k, v in kw.items():
            if v:
                cx[k] = v
        return cx

    def get(self, request, *args, **kwargs):
        user = request.user
        if not user.is_authenticated:
            return redirect(self.logout_view_name)
        assert isinstance(user, User)

        cx = self.get_context_data()

        try:
            self.get_session(request).send_code()
        except ValidationError as e:
            logger.info('2FA_WARNING: {} / {}'.format(user, e))
            cx['error'] = ' '.join(e.messages)

        return render(request, self.template_name, cx)

    def get_session(self, request: HttpRequest, reset: bool = False) -> TwoFactorSession:
        user, ip, user_agent, phone = self.get_session_const(request)
        ses_id = request.session.get('j2fa_session')
        ses = TwoFactorSession.objects.filter(id=ses_id).first() if ses_id else None
        assert ses is None or isinstance(ses, TwoFactorSession)
        if not ses or not ses.is_valid(user, ip, user_agent) or reset:
            since = now() - timedelta(hours=24)
            if hasattr(settings, 'MAX_FAILED_ATTEMPTS_24H'):
                max_failed_attempts_24h = settings.MAX_FAILED_ATTEMPTS_24H
            else:
                max_failed_attempts_24h = DEFAULT_MAX_FAILED_ATTEMPTS_24H

            if TwoFactorSession.objects.count_failed_attempts(user, ip, since) > max_failed_attempts_24h:
                raise TwoFactorAuthError(_('Too many attempts in 24h'))

            ses = TwoFactorSession.objects.create(user=user, ip=ip, user_agent=user_agent, phone=phone, code=j2fa_make_code())
            request.session['j2fa_session'] = ses.id
        return ses

    def get_session_const(self, request: HttpRequest):
        user = request.user
        assert isinstance(user, User)
        ip = get_real_ip(request)
        if ip is None and settings.DEBUG:
            ip = '127.0.0.1'
        user_agent = request.META['HTTP_USER_AGENT']
        phone = j2fa_phone_filter(self.get_user_phone(user))
        if not phone:
            raise TwoFactorAuthError(_('your.phone.number.missing.from.system'))
        return user, ip, user_agent, phone

    def post(self, request, *args, **kwargs):  # pylint: disable=unused-argument
        cx = self.get_context_data()
        form = cx['form']
        assert isinstance(form, TwoFactorForm)

        if form.is_valid():
            try:
                ses = self.get_session(request)
                assert isinstance(ses, TwoFactorSession)
                code = form.cleaned_data['code']
                user, ip, user_agent, phone = self.get_session_const(request)
                logger.info('2FA_POST_CODE: {} {} {} "{}" vs {}'.format(user, ip, user_agent, phone, ses.code))
                if ses.code != code:
                    self.get_session(request, reset=True).send_code()
                    raise TwoFactorAuthError(_('Invalid code, sending new one'))

                logger.info('2FA_PASS: {} / {}'.format(user, ses))
                TwoFactorSession.objects.archive_old_sessions(user, ses)

                return redirect(cx.get('next'))
            except TwoFactorAuthError as e:
                form.add_error(None, e)
            except Exception as e:
                form.add_error(None, e)

        return render(request, self.template_name, cx)



# Generated by Django 2.1 on 2018-08-07 22:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('j2fa', '0002_auto_20180209_0241'),
    ]

    operations = [
        migrations.AlterField(
            model_name='twofactorsession',
            name='active',
            field=models.BooleanField(blank=True, db_index=True, default=False),
        ),
        migrations.AlterField(
            model_name='twofactorsession',
            name='archived',
            field=models.BooleanField(blank=True, db_index=True, default=False),
        ),
    ]



# Generated by Django 2.0 on 2018-02-09 02:41

from django.db import migrations, models
import j2fa.helpers


class Migration(migrations.Migration):

    dependencies = [
        ('j2fa', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='twofactorsession',
            name='code',
            field=models.CharField(blank=True, default=j2fa.helpers.j2fa_make_code, max_length=8),
        ),
    ]



#pylint: disable=logging-format-interpolation
import logging
from django.conf import settings
from ipware.ip import get_real_ip
from j2fa.models import TwoFactorSession
from django.contrib.auth.models import User
from django.contrib.sessions.backends.base import SessionBase
from django.http import HttpRequest
from django.shortcuts import redirect
from django.urls import resolve, reverse


# these can be overriden in settings.py by defining variable J2FA_EXCLUDED_ROUTES
J2FA_EXCLUDED_ROUTES = [
    'j2fa-obtain-auth',
    'logout',
]

logger = logging.getLogger(__name__)


class Ensure2FactorAuthenticatedMiddleware:
    """
    Ensures that User is 2-factor authenticated.
    Place after session init.
    """
    def __init__(self, get_response=None):
        self.get_response = get_response

    def is_2fa_required(self, user: User):
        return user.is_authenticated and hasattr(user, 'profile') and user.profile.require_2fa

    def __call__(self, request: HttpRequest):
        user = request.user
        if isinstance(user, User):
            session = request.session
            assert isinstance(session, SessionBase)
            user_agent = request.META.get('HTTP_USER_AGENT', '')
            ip = get_real_ip(request)
            if ip is None and settings.DEBUG:
                ip = '127.0.0.1'

            # excluded routes
            excluded_routes = J2FA_EXCLUDED_ROUTES
            if hasattr(settings, 'J2FA_EXCLUDED_ROUTES'):
                excluded_routes = getattr(settings, 'J2FA_EXCLUDED_ROUTES')

            # Load auth data, and redirect if not valid
            route_name = resolve(request.path_info).url_name
            if self.is_2fa_required(user) and route_name not in excluded_routes:
                j2fa_session_id = session.get('j2fa_session', None)
                j2fa_session = TwoFactorSession.objects.filter(id=j2fa_session_id).first()
                assert j2fa_session is None or isinstance(j2fa_session, TwoFactorSession)
                if not j2fa_session or not j2fa_session.is_valid(user, ip, user_agent) or not j2fa_session.active:
                    logger.info('J2FA_REQUIRE_AUTH: route={}, j2fa_session_id={}'.format(route_name, j2fa_session_id))
                    return redirect(reverse('j2fa-obtain-auth'))

        # get response
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        return response



from django import forms
from django.utils.translation import gettext_lazy as _


class TwoFactorForm(forms.Form):
    code = forms.CharField(label=_('two.factor.code.label'), max_length=8, min_length=1)



#pylint: disable=logging-format-interpolation
import logging
from datetime import timedelta
from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Q
from django.utils.timezone import now
from j2fa.helpers import j2fa_make_code, j2fa_send_sms


logger = logging.getLogger(__name__)

A2FA_SESSION_EXPIRES_MINUTES = 60*24
A2FA_SESSION_CLEANUP_DAYS = 7


class TwoFactorSessionManager(models.Manager):
    def count_failed_attempts(self, user, ip, since):
        return self.filter(Q(ip=ip) | Q(user=user)).filter(created__gt=since, archived=False).count()

    def archive_old_sessions(self, user: User, current):
        self.filter(user=user).exclude(id=current.id).update(archived=True, active=False)
        current.active = True
        current.archived = False
        current.save()

    def clean_up_old_sessions(self):
        old = now() - timedelta(days=A2FA_SESSION_CLEANUP_DAYS)
        self.filter(created__lt=old).delete()


class TwoFactorSession(models.Model):
    objects = TwoFactorSessionManager()
    created = models.DateTimeField(default=now, db_index=True, blank=True)
    user = models.ForeignKey(User, related_name='+', on_delete=models.CASCADE)
    user_agent = models.CharField(max_length=512)
    ip = models.GenericIPAddressField(db_index=True)
    phone = models.CharField(max_length=32, db_index=True)
    code = models.CharField(max_length=8, default=j2fa_make_code, blank=True)
    active = models.BooleanField(default=False, db_index=True, blank=True)
    archived = models.BooleanField(default=False, db_index=True, blank=True)

    def __str__(self):
        return '[{}]'.format(self.id)

    def is_valid(self, user: User, ip: str, user_agent: str) -> bool:
        if now() - self.created > timedelta(minutes=A2FA_SESSION_EXPIRES_MINUTES):
            return False
        return self.user == user and self.ip == ip and self.user_agent == user_agent

    def send_code(self):
        logger.info('2FA_SEND_CODE: {} / {} / {}'.format(self.user, self.phone, self.code))
        if settings.SMS_TOKEN:
            j2fa_send_sms(self.phone, self.code)
